//  Copyright © Microsoft Corporation
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

package org.opengroup.osdu.register.provider.azure.di;

import com.azure.identity.DefaultAzureCredentialBuilder;
import com.azure.security.keyvault.keys.cryptography.CryptographyClient;
import com.azure.security.keyvault.keys.cryptography.CryptographyClientBuilder;
import com.azure.security.keyvault.secrets.SecretClient;
import com.azure.security.keyvault.secrets.models.KeyVaultSecret;
import com.microsoft.azure.AzureEnvironment;
import com.microsoft.azure.credentials.ApplicationTokenCredentials;
import com.microsoft.azure.credentials.AzureTokenCredentials;
import com.microsoft.azure.management.eventgrid.v2019_01_01.implementation.EventGridManager;
import com.microsoft.rest.LogLevel;
import lombok.Getter;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

import javax.inject.Named;
import java.util.HashMap;

@Configuration
@Getter
public class AzureBootstrapConfig {
    @Value("${azure.keyvault.url}")
    private String keyVaultURL;

    @Value("${azure.cosmosdb.database}")
    private String cosmosDBName;

    @Value("${azure.cryptoKey.identifier}")
    private String keyIdentifier;

    @Value("${azure.clientId}")
    private String azureClientId;

    @Value("${azure.clientSecret}")
    private String azureClientSecret;

    @Value("${azure.tenantId}")
    private String azureTenantId;

    @Value("${azure.subscriptionId}")
    private String azureSubscriptionId;

    @Value("${azure.resourceGroupName}")
    private String resourceGroupName;

    @Bean
    @Named("KEY_VAULT_URL")
    public String keyVaultURL() {
        return keyVaultURL;
    }

    @Bean
    @Named("COSMOS_DB_NAME")
    public String cosmosDBName() {
        return cosmosDBName;
    }

    @Bean
    @Named("COSMOS_ENDPOINT")
    public String cosmosEndpoint(SecretClient kv) {
        return getKeyVaultSecret(kv, "cosmosdb-uri");
    }

    @Bean
    @Named("COSMOS_KEY")
    public String cosmosKey(SecretClient kv) {
        return getKeyVaultSecret(kv, "cosmos-db-key");
    }

    @Bean
    public CryptographyClient getCryptographyClient() {
        return new CryptographyClientBuilder()
                .keyIdentifier(keyIdentifier)
                .credential(new DefaultAzureCredentialBuilder().build())
                .buildClient();
    }

    @Bean
    public EventGridManager eventGridManager() {
        AzureTokenCredentials azureTokenCredentials = getAzureTokenCredentials();
        return EventGridManager
                .configure()
                .withLogLevel(LogLevel.BASIC)
                .authenticate(azureTokenCredentials, azureTokenCredentials.defaultSubscriptionId());
    }


    private AzureTokenCredentials getAzureTokenCredentials() {
        AzureEnvironment azureEnvironment = new AzureEnvironment(new HashMap<>());
        azureEnvironment.endpoints().putAll(AzureEnvironment.AZURE.endpoints());
        return new ApplicationTokenCredentials(
                azureClientId,
                azureTenantId,
                azureClientSecret,
                azureEnvironment).withDefaultSubscriptionId(azureSubscriptionId);
    }

    String getKeyVaultSecret(SecretClient kv, String secretName) {
        KeyVaultSecret secret = kv.getSecret(secretName);
        if (secret == null) {
            throw new IllegalStateException(String.format("No secret found with name %s", secretName));
        }

        String secretValue = secret.getValue();
        if (secretValue == null) {
            throw new IllegalStateException(String.format(
                    "Secret unexpectedly missing from KeyVault response for secret with name %s", secretName));
        }

        return secretValue;
    }
}
