// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import com.azure.cosmos.FeedOptions;
import com.azure.cosmos.SqlParameter;
import com.azure.cosmos.SqlParameterList;
import com.azure.cosmos.SqlQuerySpec;
import com.google.cloud.Timestamp;
import org.opengroup.osdu.azure.CosmosStore;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.di.CosmosContainerConfig;
import org.opengroup.osdu.register.provider.azure.util.CryptographyUtil;
import org.opengroup.osdu.register.provider.interfaces.subscriber.ISubscriptionRepository;
import org.opengroup.osdu.register.subscriber.model.GsaSecret;
import org.opengroup.osdu.register.subscriber.model.GsaSecretValue;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.List;
import java.util.Optional;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

@Repository
public class SubscriptionRepository implements ISubscriptionRepository {

    private final ReentrantLock mutex = new ReentrantLock();

    @Autowired
    private AzureBootstrapConfig azureBootstrapConfig;

    @Autowired
    private CosmosContainerConfig cosmosContainerConfig;

    @Autowired
    private CosmosStore cosmosStore;

    @Autowired
    private DpsHeaders dpsHeaders;

    @Autowired
    private JaxRsDpsLog logger;

    @Autowired
    private CryptographyUtil cryptographyUtil;

    @Autowired
    private PushSubscription pushSubscription;

    @Override
    public Subscription create(Subscription input) {

        try {
            mutex.lock();

            SubscriptionDoc doc = new SubscriptionDoc(input, dpsHeaders.getPartitionId());
            String encryptedSecret = cryptographyUtil.encryptData(input.getSecret().toString());
            doc.setSecretValue(encryptedSecret);
            doc.setKeyName(cryptographyUtil.getKeyName());
            doc.setKeyVersion(cryptographyUtil.getKeyVersion());

            try {
                cosmosStore.createItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), doc);
            }
            catch (AppException e) {
                if(e.getError().getCode() == 409) {
                    logger.error("A subscriber already exists with the same topic and endpoint combination");
                    throw new AppException(409, "Conflict", "A subscriber already exists with the same topic and endpoint combination");
                }
                else {
                    logger.error(e.getMessage());
                    throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
                }
            }

            boolean isSubscriptionCreated = pushSubscription.createPushSubscription(input);

            if(isSubscriptionCreated) {
                return input;
            }
            else {
                cosmosStore.deleteItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), input.getId(), dpsHeaders.getPartitionId());
                logger.error("Unexpected error creating subscription");
                throw new AppException(500, "Server Error", "Unexpected error creating subscription");
            }
        }
        finally {
            mutex.unlock();
        }

    }

    @Override
    public Subscription get(String id) {

        SubscriptionDoc doc = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId(), SubscriptionDoc.class)
                .orElseThrow(() -> {
                    logger.error(String.format("Subscriber with id %s does not exist.", id));
                    return new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", id));
                });

        return convertToSubscriptionClass(doc);
    }

    @Override
    public List<Subscription> query(String notificationId) {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.notificationId = @notificationId and c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@notificationId", notificationId));
        pars.add(new SqlParameter("@partitionId", dpsHeaders.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());
    }

    @Override
    public boolean delete(String id) {

        try {
            mutex.lock();

            Optional<SubscriptionDoc> subscription = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId(), SubscriptionDoc.class);

            if(!subscription.isPresent()){
                return false;
            }

            boolean isDeleted = pushSubscription.deletePushSubscription(subscription.get().getNotificationId(), subscription.get().getTopic());

            if(isDeleted) {
                cosmosStore.deleteItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId());
                return true;
            }
            else {
                logger.error("Unexpected error deleting subscription");
                throw new AppException(500, "Server Error", "Unexpected error deleting subscription");
            }
        }

        finally {
            mutex.unlock();
        }

    }

    @Override
    public boolean patch(Subscription subscription, Secret secret){
        return updateSecret(subscription, secret);
    }

    @Override
    public List<Subscription> getAll() {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@partitionId", dpsHeaders.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());
    }

    private boolean updateSecret(Subscription subscription, Secret secret) {
        SubscriptionDoc doc = new SubscriptionDoc(subscription, dpsHeaders.getPartitionId());
        doc.setSecretType(secret.getSecretType());
        doc.setSecretValue(cryptographyUtil.encryptData(secret.toString()));

        cosmosStore.upsertItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), doc);
        return true;
    }

    private Subscription convertToSubscriptionClass(SubscriptionDoc doc) {

        String secretValue = cryptographyUtil.decryptData(doc.getSecretValue());
        Secret secret;
        if (doc.getSecretType().equals(Constants.GSASecret)) {
            GsaSecret gsaSecret = new GsaSecret();
            String[] splitSecret = secretValue.split("`");
            gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
            secret = gsaSecret;
        } else {
            HmacSecret hmacSecret = new HmacSecret();
            hmacSecret.setValue(secretValue);
            secret = hmacSecret;
        }
        return new Subscription(doc.getId(), doc.getName(),doc.getDescription(), doc.getTopic(), doc.getPushEndpoint(),
                doc.getCreatedBy(), Timestamp.of(doc.getCreatedOnEpoch()), doc.getNotificationId(), secret);
    }
}
