// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import com.microsoft.azure.arm.model.Indexable;
import com.microsoft.azure.management.eventgrid.v2019_01_01.EventSubscription;
import com.microsoft.azure.management.eventgrid.v2019_01_01.EventSubscriptions;
import com.microsoft.azure.management.eventgrid.v2019_01_01.implementation.EventGridManager;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.AppServiceConfig;
import rx.Completable;
import rx.Observable;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
public class PushSubscriptionTest {

    private static final String pushEndpoint = "http://testing-env.com/api/v1/records-changed";
    private static final String subscriptionId = "subscription-id";
    private static final String resourceGroupName = "resource-group";
    private static final String topicName = "topic-name";
    private static final String errorMessage = "some error message";
    private static final String scope = String.format("/subscriptions/%s/resourceGroups/%s/providers/Microsoft.EventGrid/topics/%s", subscriptionId, resourceGroupName, topicName);
    private final Observable<Indexable> observable = Observable.from(new Indexable[]{});
    private final Observable<Indexable> observableError = Observable.error(new Throwable(errorMessage));
    private final Completable completable = Completable.complete();
    private final Completable completableError = Completable.error(new Throwable(errorMessage));
    @Mock
    private EventGridManager eventGridManager;

    @Mock
    private AzureBootstrapConfig azureBootstrapConfig;

    @Mock
    private JaxRsDpsLog logger;

    @Mock
    private Subscription subscription;

    @Mock
    private EventSubscriptions eventSubscriptions;

    @Mock
    private EventSubscription.DefinitionStages.Blank eventSubscription;

    @Mock
    private EventSubscription.DefinitionStages.WithCreate eventSubscriptionWithCreate;

    @Mock
    private AppServiceConfig serviceConfig;

    @InjectMocks
    private PushSubscription pushSubscription;

    @BeforeEach
    public void init() {
        lenient().when(azureBootstrapConfig.getResourceGroupName()).thenReturn(resourceGroupName);
        lenient().when(azureBootstrapConfig.getAzureSubscriptionId()).thenReturn(subscriptionId);
        lenient().when(subscription.getNotificationId()).thenReturn(subscriptionId);
        lenient().when(subscription.getTopic()).thenReturn(topicName);
        lenient().when(serviceConfig.getRecordsChangePubsubEndpoint()).thenReturn(pushEndpoint);
        lenient().when(eventGridManager.eventSubscriptions()).thenReturn(eventSubscriptions);
        lenient().when(eventSubscriptions.define(subscriptionId)).thenReturn(eventSubscription);
        lenient().when(eventSubscription.withScope(scope)).thenReturn(eventSubscriptionWithCreate);
        lenient().when(eventSubscriptionWithCreate.withDestination(any())).thenReturn(eventSubscriptionWithCreate);
    }

    @Test
    public void shouldReturnTrueWhenPushSubscriptionCreated() {

        when(eventSubscriptionWithCreate.createAsync()).thenReturn(observable);
        boolean result = pushSubscription.createPushSubscription(subscription);

        assertTrue(result);
        verify(azureBootstrapConfig, times(1)).getResourceGroupName();
        verify(azureBootstrapConfig, times(1)).getAzureSubscriptionId();
        verify(subscription, times(1)).getNotificationId();
        verify(serviceConfig, times(1)).getRecordsChangePubsubEndpoint();
        verify(subscription, times(1)).getTopic();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(eventSubscriptions, times(1)).define(subscriptionId);
        verify(eventSubscription, times(1)).withScope(scope);
        verify(logger, times(1)).info(String.format("Subscription with id %s created successfully", subscriptionId));
    }

    @Test
    public void shouldReturnFalseWhenPushSubscriptionCreationFailed() {

        when(eventSubscriptionWithCreate.createAsync()).thenReturn(observableError);
        boolean result = pushSubscription.createPushSubscription(subscription);

        assertFalse(result);
        verify(azureBootstrapConfig, times(1)).getResourceGroupName();
        verify(azureBootstrapConfig, times(1)).getAzureSubscriptionId();
        verify(subscription, times(1)).getNotificationId();
        verify(subscription, times(1)).getTopic();
        verify(serviceConfig, times(1)).getRecordsChangePubsubEndpoint();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(eventSubscriptions, times(1)).define(subscriptionId);
        verify(eventSubscription, times(1)).withScope(scope);
        verify(logger, times(1)).error(String.format("Creating subscription with id %s failed with error %s", subscriptionId, errorMessage));
    }

    @Test
    public void shouldReturnTrueWhenPushSubscriptionDeleted() {

        when(eventSubscriptions.deleteAsync(scope, subscriptionId)).thenReturn(completable);

        boolean result = pushSubscription.deletePushSubscription(subscriptionId, topicName);

        assertTrue(result);
        verify(azureBootstrapConfig, times(1)).getResourceGroupName();
        verify(azureBootstrapConfig, times(1)).getAzureSubscriptionId();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(logger, times(1)).info(String.format("Subscription with id %s deleted successfully", subscriptionId));
    }

    @Test
    public void shouldReturnFalseWhenPushSubscriptionDeletionFailed() {

        when(eventSubscriptions.deleteAsync(scope, subscriptionId)).thenReturn(completableError);

        boolean result = pushSubscription.deletePushSubscription(subscriptionId, topicName);

        assertFalse(result);
        verify(azureBootstrapConfig, times(1)).getResourceGroupName();
        verify(azureBootstrapConfig, times(1)).getAzureSubscriptionId();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(logger, times(1)).error(String.format("Delete operation on subscription with id %s failed with error %s", subscriptionId, errorMessage));
    }
}
